from ud4d.logger import logger
from ud4d.docker_manager import DeviceContainer, DeviceContainerManager


class UEvent(object):
    """ convert udev info (str) into object """
    def __init__(self, udev_info_list):
        for each_arg in udev_info_list[1:]:
            name, value = each_arg.split('=')
            self.__dict__[name] = value

    def is_android(self) -> bool:
        return 'ID_SERIAL_SHORT' in self.__dict__

    def get_serial_no(self) -> str:
        assert self.is_android(), 'device not android'
        return self.ID_SERIAL_SHORT

    def get_dev_name(self) -> str:
        if hasattr(self, 'DEVNAME'):
            return self.DEVNAME
        logger.info('no device name found, return device path instand')
        return self.DEVPATH

    def get_event_id(self) -> str:
        return self.USEC_INITIALIZED

    def get_action_name(self) -> str:
        return self.ACTION


class UEventManager(object):
    """ manage events about android, ignore everything else """
    _event_dict = dict()

    @classmethod
    def add_event(cls, uevent: UEvent):
        event_id = uevent.get_event_id()
        action = uevent.get_action_name()

        # only 'bind' and 'unbind'
        if action == 'bind' and uevent.is_android():
            cls._event_dict[event_id] = uevent
            logger.info('event id [%s] added', event_id)

            # build docker container
            container = DeviceContainer(uevent.get_serial_no(), uevent.get_dev_name())
            DeviceContainerManager.add(container)

        elif action == 'unbind' and event_id in cls._event_dict:
            # remove docker container
            DeviceContainerManager.remove(cls._event_dict[event_id].get_serial_no())

            del cls._event_dict[event_id]
            logger.info('event id [%s] ended', event_id)

        else:
            # ignore other events
            pass

    @classmethod
    def get_event_dict(cls):
        return cls._event_dict



import docker

from ud4d import config as u_config
from ud4d.logger import logger


# device image name
device_image_name = u_config.DEVICE_IMAGE_NAME
logger.info('device image name: %s', device_image_name)

# docker client
client = docker.from_env()

# check
try:
    device_image = client.images.get(device_image_name)
except docker.errors.ImageNotFound as e:
    logger.error('device image not found: %s, you should build it first', device_image_name)
    raise e


class DeviceContainer(object):
    """ android device container object """
    def __init__(self, serial_no, device_name):
        self.serial_no = serial_no
        self.device_name = device_name
        self.container = client.containers.run(
            device_image,
            detach=True,
            # container name
            name='android_device_' + serial_no,
            # bind devices
            devices=['{}:{}:rwm'.format(device_name, device_name)],
            # keep it running
            tty=True,
        )


class DeviceContainerManager(object):
    _container_dict = dict()

    @classmethod
    def add(cls, device_container: DeviceContainer):
        cls._container_dict[device_container.serial_no] = device_container
        logger.info('device container [%s] running', device_container.serial_no)

    @classmethod
    def remove(cls, serial_no):
        if serial_no in cls._container_dict:
            cls._container_dict[serial_no].container.remove(force=True)
            del cls._container_dict[serial_no]
            logger.info('device container [%s] removed', serial_no)
        else:
            logger.warn('device container [%s] not existed', serial_no)

    @classmethod
    def remove_all(cls):
        """ remove all containers """
        for each_serial_no in cls._container_dict:
            cls.remove(each_serial_no)



import subprocess


# check: udevadm is runnable?
assert not subprocess.check_call(['udevadm', '--version']), 'udevadm started failed. Is your system linux?'

CHARSET = 'utf-8'
UDEVADM_PATH = subprocess.check_output(['which', 'udevadm']).decode(CHARSET).strip()
DEVICE_IMAGE_NAME = 'williamfzc/adb_tester:0.1'



import subprocess
import time

from ud4d import config as u_config
from ud4d.logger import logger


class UDevDetector(object):
    """ udevadm process management """
    _process = None

    @classmethod
    def start(cls):
        cls._process = subprocess.Popen(
            [u_config.UDEVADM_PATH, 'monitor', '-u', '--subsystem-match=usb', '--environment'],
            stdout=subprocess.PIPE,
        )
        time.sleep(1)
        assert cls.is_running(), 'udevadm not running'
        logger.info('udevadm process up')

        # ignore unused header
        cls.read_event()

    @classmethod
    def stop(cls):
        if cls._process and cls.is_running():
            cls._process.kill()
            cls._process = None
            logger.info('udevadm process down')
        else:
            logger.warn('udevadm process already down without killing')

    @classmethod
    def is_running(cls) -> bool:
        return not bool(cls._process.poll())

    @classmethod
    def read_line(cls) -> str:
        return cls._process.stdout.readline().decode(u_config.CHARSET)

    @classmethod
    def read_event(cls) -> list:
        event_content = []
        new_line = cls.read_line().strip()
        while new_line:
            event_content.append(new_line)
            new_line = cls.read_line().strip()
        return event_content



import contextlib
import threading

from ud4d.detector import UDevDetector
from ud4d.docker_manager import DeviceContainerManager
from ud4d.logger import logger
from ud4d.event import UEventManager, UEvent


UD4D_STATUS = False


@contextlib.contextmanager
def safe_ud4d():
    UDevDetector.start()
    logger.info('udev detector started')
    yield UDevDetector
    DeviceContainerManager.remove_all()
    logger.info('all device container removed')
    UDevDetector.stop()
    logger.info('udev detector stopped')


def start_ud4d():
    global UD4D_STATUS
    UD4D_STATUS = True

    def inner():
        UDevDetector.start()
        while UD4D_STATUS:
            first_event = UDevDetector.read_event()
            event_object = UEvent(first_event)
            UEventManager.add_event(event_object)
        UDevDetector.stop()

    threading.Thread(target=inner).start()


def stop_ud4d():
    global UD4D_STATUS
    UD4D_STATUS = False



from setuptools import setup, find_packages


setup(
    name='ud4d',
    version='0.1.0',
    description='USB Device Detector for Docker usage. Support Linux only.',
    author='williamfzc',
    author_email='fengzc@vip.qq.com',
    url='https://github.com/doringland/ud4d',
    packages=find_packages(),
    install_requires=[
        'logzero',
        'docker',
    ]
)
